open Arch
open Ids
open Exprs

module type Inst_t = sig
  module A: Vec_arch_t

  type inst_type = Vec | Scal [@@deriving eq, show]

  type t =
    | Nop
    (* Allocated is a placeholder for "factorized out": instructions that are
     * in some way "out of scope", we don't expect to generate a fresh name for
     * them. Still uneasy with that though *)
    | Allocated of inst_type * string
    | Assign of t * inst_type * string
    | Read of Tensor.t * Tensor.accesses
    | Add of t * t
    | Sub of t * t
    | Mul of t * t
    | Write of t * Tensor.t * Tensor.accesses
    | Vread of Tensor.t * Tensor.accesses
    | Vadd of t * t
    | Vsub of t * t
    | Vmul of t * t
    | Vbcst of t
    | Vwrite of t * Tensor.t * Tensor.accesses [@@deriving eq, show]

  val vectorize_on_dim: Dim.id -> t -> t
  val map_expr: Dim.id -> (VarExpr.t -> VarExpr.t ) -> t -> t
  val access: Dim.id -> t -> bool
  val is_tensor_readonly: t list -> Tensor.t -> bool

  type mem_accesses_inf = (Tensor.t * Tensor.accesses) * (inst_type * string)
  val fact_loads_on_dims:  Dim.id list -> mem_accesses_inf list  -> t list
    -> mem_accesses_inf list * t list * t list
  val fact_stores_on_dims:  Dim.id list -> mem_accesses_inf list  -> t list
    -> mem_accesses_inf list * t list * t list
  val swap_tensor: Tensor.t -> Tensor.t -> t list -> t list
  val swap_accesses: (Tensor.t -> Tensor.accesses -> Tensor.accesses) -> t list -> t list
  val gen_code: string -> t list -> string list * string list * string
end

module I(A:Vec_arch_t) : Inst_t with module A = A
